package djmil.cordacheckers.cordaclient;

import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.Assertions.assertThatThrownBy;

import java.util.ArrayList;
import java.util.Arrays;

import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;

import djmil.cordacheckers.cordaclient.dao.GameState;
import djmil.cordacheckers.cordaclient.dao.GameState.Status;
import djmil.cordacheckers.cordaclient.dao.Stone;
import djmil.cordacheckers.user.HoldingIdentityResolver;

@SpringBootTest
public class GameBoardTests {
    @Autowired
    CordaClient cordaClient;

    @Autowired
    HoldingIdentityResolver holdingIdentityResolver;

    final String whitePlayerName = "alice";
    final String blackPlayerName = "bob";

    final static Stone WHITE_MAN  = new Stone(Stone.Color.WHITE, Stone.Type.MAN);
    final static Stone WHITE_KING = new Stone(Stone.Color.WHITE, Stone.Type.KING);
    final static Stone BLACK_MAN  = new Stone(Stone.Color.BLACK, Stone.Type.MAN);
    final static Stone BLACK_KING = new Stone(Stone.Color.BLACK, Stone.Type.KING);

    @Test
    void testSurrender() {
        final var hiWhite  = holdingIdentityResolver.getByUsername(whitePlayerName);
        final var hiBlack = holdingIdentityResolver.getByUsername(blackPlayerName);
        final String message = "GameBoard SURRENDER test";

        final GameState game = cordaClient.gameProposalCreate(
            hiWhite, hiBlack, Stone.Color.BLACK, message);

        System.out.println("Game UUID " +game.uuid());

        final GameState acceptedGameBlackView = cordaClient.gameProposalAccept(
            hiBlack, game.uuid());

        assertThat(acceptedGameBlackView.opponentColor()).isEqualByComparingTo(Stone.Color.WHITE);
        assertThat(acceptedGameBlackView.status()).isEqualByComparingTo(Status.GAME_BOARD_WAIT_FOR_OPPONENT);

        assertThatThrownBy(() -> { // Black can not surrender, since it is opponent's turn
            cordaClient.gameBoardSurrender(
                hiBlack, game.uuid());
        });

        final GameState surrendererGameView = cordaClient.gameBoardSurrender(
            hiWhite, game.uuid());

        assertThat(surrendererGameView.opponentName()).isEqualToIgnoringCase(blackPlayerName);
        assertThat(surrendererGameView.opponentColor()).isEqualByComparingTo(Stone.Color.BLACK);
        assertThat(surrendererGameView.status()).isEqualByComparingTo(Status.GAME_RESULT_YOU_LOOSE);

        final GameState winnerGameView = cordaClient.gameStateGet(
            hiBlack, game.uuid());

        assertThat(winnerGameView.opponentName()).isEqualToIgnoringCase(whitePlayerName);
        assertThat(winnerGameView.opponentColor()).isEqualByComparingTo(Stone.Color.WHITE);
        assertThat(winnerGameView.status()).isEqualByComparingTo(Status.GAME_RESULT_YOU_WON);
    }

    @Test
    void testMove() {
        final var hiWhite  = holdingIdentityResolver.getByUsername(whitePlayerName);
        final var hiBlack = holdingIdentityResolver.getByUsername(blackPlayerName);
        final String message = "GameBoard MOVE test";

        final GameState game = cordaClient.gameProposalCreate(
            hiWhite, hiBlack, Stone.Color.BLACK, message);
        System.out.println("Game UUID " +game.uuid());

        final var m0 = cordaClient.gameProposalAccept(hiBlack, game.uuid());

        assertThatThrownBy(() -> { 
            cordaClient.gameBoardMove(hiBlack, game.uuid(), move(12, 16), 
                "Black can not move, since it is opponent's turn");
        });

        assertThatThrownBy(() -> { 
            cordaClient.gameBoardMove(hiWhite, game.uuid(), move(17, 14),
            "Trying to move an empty tile");
        });

        final var m1 = cordaClient.gameBoardMove(hiWhite, game.uuid(), move(22, 18), null);
        assertThat(m0.board().get(22)).isEqualTo(WHITE_MAN);
        assertThat(m0.board().get(18)).isNull();
        assertThat(m1.board().get(22)).isNull();
        assertThat(m1.board().get(18)).isEqualTo(WHITE_MAN);
        assertThat(m1.status()).isEqualByComparingTo(Status.GAME_BOARD_WAIT_FOR_OPPONENT);
    }

    ArrayList<Integer> move(int from, int to) {
        return new ArrayList<Integer>(Arrays.asList(from, to));
    }
}
