package djmil.cordacheckers.contracts;

import static djmil.cordacheckers.contracts.GameCommand.requireThat;
import static djmil.cordacheckers.contracts.UtxoLedgerTransactionUtil.getSingleCommand;

import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import djmil.cordacheckers.states.Stone;
import djmil.cordacheckers.states.Stone.Color;
import net.corda.v5.ledger.utxo.transaction.UtxoLedgerTransaction;

public class GameBoardContract implements net.corda.v5.ledger.utxo.Contract {
    
    public static class MoveResult {
        final public Map<Integer, Stone> board;
        final public Stone.Color moveColor;
        
        public MoveResult(Map<Integer, Stone> board, Color moveColor) {
            this.board = board;
            this.moveColor = moveColor;
        }

        public static class Exception extends RuntimeException {
            public Exception(String message) {
                super(message);
            }
        }
    }

    private final static Logger log = LoggerFactory.getLogger(GameBoardContract.class);

    @Override
    public void verify(UtxoLedgerTransaction trx) {
        log.info("GameBoardContract.verify() called");

        requireThat(trx.getCommands().size() == 1, GameCommand.REQUIRE_SINGLE_COMMAND);
        final GameCommand command = getSingleCommand(trx, GameCommand.class);

        switch (command.getAction()) {
            case GAME_PROPOSAL_ACCEPT:
                command.validateGameProposalAccept(trx);
                break;
        
            case GAME_BOARD_MOVE:
                command.validateGameBoardMove(trx, command.getMove());
                break;

            case GAME_BOARD_SURRENDER:
                command.validateGameBoardSurrender(trx);
                break;

            default:
                throw new GameCommand.ActionException();
        }
    }

    public static MoveResult applyMove(List<Integer> move, Map<Integer, Stone> board, Stone.Color moveColor) {
        final int mFrom = move.get(0);
        final int mTo   = move.get(1);

        final Stone piece = board.get(mFrom);
        if (piece == null)
            throw new MoveResult.Exception("An empty starting tile");

        if (piece.getColor() != moveColor)
            throw new MoveResult.Exception("Can not move opponent's piece");

        if (board.get(mTo) != null)
            throw new MoveResult.Exception("An occupied finishing tile");

        final Map<Integer, Stone> newBoard = new LinkedHashMap<Integer, Stone>(board);
        newBoard.remove(mFrom);
        newBoard.put(mTo, piece);

        return new GameBoardContract.MoveResult(newBoard, moveColor.opposite());
    }

    public final static Map<Integer, Stone> initialBoard =  Map.ofEntries(
        // Inspired by Checkers notation rules: https://www.bobnewell.net/nucleus/checkers.php
        Map.entry( 1, Stone.BLACK_MAN),
        Map.entry( 2, Stone.BLACK_MAN),
        Map.entry( 3, Stone.BLACK_MAN),
        Map.entry( 4, Stone.BLACK_MAN),
        Map.entry( 5, Stone.BLACK_MAN),
        Map.entry( 6, Stone.BLACK_MAN),
        Map.entry( 7, Stone.BLACK_MAN),
        Map.entry( 8, Stone.BLACK_MAN),
        Map.entry( 9, Stone.BLACK_MAN),
        Map.entry(10, Stone.BLACK_MAN),
        Map.entry(11, Stone.BLACK_MAN),
        Map.entry(12, Stone.BLACK_MAN),

        Map.entry(21, Stone.WHITE_MAN),
        Map.entry(22, Stone.WHITE_MAN),
        Map.entry(23, Stone.WHITE_MAN),
        Map.entry(24, Stone.WHITE_MAN),
        Map.entry(25, Stone.WHITE_MAN),
        Map.entry(26, Stone.WHITE_MAN),
        Map.entry(27, Stone.WHITE_MAN),
        Map.entry(28, Stone.WHITE_MAN),
        Map.entry(29, Stone.WHITE_MAN),
        Map.entry(30, Stone.WHITE_MAN),
        Map.entry(31, Stone.WHITE_MAN),
        Map.entry(32, Stone.WHITE_MAN)
        );


}
