package djmil.cordacheckers.gamestate;

import static djmil.cordacheckers.contracts.UtxoLedgerTransactionUtil.getSingleCommand;
import static djmil.cordacheckers.contracts.UtxoLedgerTransactionUtil.getSingleOutputState;
import static djmil.cordacheckers.contracts.UtxoLedgerTransactionUtil.getSingleInputState;

import djmil.cordacheckers.contracts.GameCommand;
import djmil.cordacheckers.states.GameBoardState;
import djmil.cordacheckers.states.GameProposalState;
import djmil.cordacheckers.states.GameResultState;
import djmil.cordacheckers.states.GameState;
import net.corda.v5.application.flows.CordaInject;
import net.corda.v5.application.flows.SubFlow;
import net.corda.v5.application.membership.MemberLookup;
import net.corda.v5.base.annotations.Suspendable;
import net.corda.v5.base.types.MemberX500Name;
import net.corda.v5.crypto.SecureHash;
import net.corda.v5.ledger.utxo.UtxoLedgerService;
import net.corda.v5.ledger.utxo.transaction.UtxoLedgerTransaction;

/*
 * A view on a ledger transaction that has GameState involved in it, from perspective of a concrete player
 */ 
public class ViewBuilder implements SubFlow<View> {
    private final SecureHash gameStateUtxoTrxId;

    public ViewBuilder (SecureHash gameStateUtxoTrxId) {
        this.gameStateUtxoTrxId = gameStateUtxoTrxId;
    }

    @CordaInject
    public UtxoLedgerService utxoLedgerService;

    @CordaInject
    public MemberLookup memberLookup;

    @Override
    @Suspendable
    public View call() {
        final UtxoLedgerTransaction gameStateUtxo = this.utxoLedgerService
            .findLedgerTransaction(gameStateUtxoTrxId);

        return buildGameStateView(gameStateUtxo);
    }

    @Suspendable
    View buildGameStateView(UtxoLedgerTransaction gameStateUtxo) {
        MemberX500Name myName = memberLookup.myInfo().getName();

        final GameCommand command = getSingleCommand(gameStateUtxo, GameCommand.class);
        final GameState state = getLatestGameStateFromTransaction(gameStateUtxo, command);
        final View.Status viewStatus = action2status(command, state, myName);

        switch (command.getAction()) {
            case GAME_PROPOSAL_CREATE:
            case GAME_PROPOSAL_CANCEL:
            case GAME_PROPOSAL_REJECT:
                if (state instanceof GameProposalState)
                    return new View(viewStatus, (GameProposalState)state, myName);
                break;

            case GAME_PROPOSAL_ACCEPT:
            case GAME_BOARD_MOVE:
                if (state instanceof GameBoardState)
                    return new View(viewStatus, (GameBoardState)state, command.getMove(), myName);
                break;
            
            case GAME_BOARD_SURRENDER:
            case GAME_RESULT_CREATE:
                if (state instanceof GameResultState)
                    return new View(viewStatus, (GameResultState)state, myName);
                break;
        }

        throw new GameCommand.ActionException();
    }

    /**
     * 
     * @param gameStateTransaction an utxo ledger transaction, that involves GameState
     * @param command
     * @return the most recent (from perspective of building a GameView) GameState for a given transaction
     * 
     * @see djmil.cordacheckers.gamestate.CommitSubFlowResponder#getActualGameStateFromTransaction(UtxoLedgerTransaction, GameCommand)
     */
    @Suspendable
    GameState getLatestGameStateFromTransaction(UtxoLedgerTransaction gameStateTransaction, GameCommand command) {
        switch (command.getAction()) {
            case GAME_PROPOSAL_CREATE:
                return getSingleOutputState(gameStateTransaction, GameProposalState.class);
            
            case GAME_PROPOSAL_REJECT:
            case GAME_PROPOSAL_CANCEL:
                return getSingleInputState(gameStateTransaction, GameProposalState.class);

            case GAME_PROPOSAL_ACCEPT:
            case GAME_BOARD_MOVE:
                return getSingleOutputState(gameStateTransaction, GameBoardState.class);
            
            case GAME_BOARD_SURRENDER:
            case GAME_RESULT_CREATE:
                return getSingleOutputState(gameStateTransaction, GameResultState.class);
        }

        throw new GameCommand.ActionException();
    }

    @Suspendable
    View.Status action2status(GameCommand command, GameState state, MemberX500Name myName) {
        final boolean myAction = command.getInitiator(state).compareTo(myName) == 0;

        switch (command.getAction()) {
            case GAME_PROPOSAL_CREATE:
                if (myAction)
                    return View.Status.GAME_PROPOSAL_WAIT_FOR_OPPONENT;
                else
                    return View.Status.GAME_PROPOSAL_WAIT_FOR_YOU;
                    
            case GAME_PROPOSAL_REJECT:
                return View.Status.GAME_PROPOSAL_REJECTED;

            case GAME_PROPOSAL_CANCEL:
                return View.Status.GAME_PROPOSAL_CANCELED;
            
            case GAME_PROPOSAL_ACCEPT:
            case GAME_BOARD_MOVE:
                if (myAction)
                    return View.Status.GAME_BOARD_WAIT_FOR_YOU;
                else
                    return View.Status.GAME_BOARD_WAIT_FOR_OPPONENT;

            case GAME_BOARD_SURRENDER:
                if (myAction)
                    return View.Status.GAME_RESULT_YOU_LOOSE;
                else
                    return View.Status.GAME_RESULT_YOU_WON; 
            
            case GAME_RESULT_CREATE:
                if (myAction)
                    return View.Status.GAME_RESULT_YOU_WON;
                else
                    return View.Status.GAME_RESULT_YOU_LOOSE; 
        }

        throw new GameCommand.ActionException();
    }
}
